#include <linux/init.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/kernel.h>
#include <asm/uaccess.h>
#include <linux/uaccess.h>
#include <linux/device.h>
#include <linux/ioport.h>
#include <asm/io.h>
#include <linux/types.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/unistd.h>
#include <linux/signal.h>
#include "X86.h"

#define IOC_MAGIC 'w' 
#define SETNUM _IOW(IOC_MAGIC, 1 , int)
#define GETNUM _IOR(IOC_MAGIC, 2 , int)
#define IOC_MAXNR 2

#define DEV_MAJOR 		152
static struct class * class_EC;
static struct device * device_EC;
//static struct device * device_EC_button;

static struct cdev *ec_cdev = NULL;
static struct cdev *ec_button_cdev = NULL;
static int adddevflg = 0;
static int devnoflg = 0;
dev_t devno;
 
int scull_major =   DEV_MAJOR;
int scull_minor =   0;

#ifndef WMIO2_EXPORTS
	#define WMIO2_EXPORTS
#endif

//#define PORT_EC_CMD    0x66
//#define PORT_EC_DATA   0x62
#define PORT_EC_CMD    0x6C
#define PORT_EC_DATA   0x68


#define  CMD_WR 		  0x81
#define  CMD_RD 		  0x80

#define MAX_BATTERY_CAPACITY 20000

static unsigned char gcData;
static bool gfDataReady=false;

bool waitForDataReady()
{
	while( !(inb(PORT_EC_CMD)&0x01) );
	return true;
}

void cleanOutputBuffer()
{
	while(inb(PORT_EC_CMD)&0x01)
	{
		inb(PORT_EC_DATA);
	};
}

bool waitForWriteFinished()
{
	while( (inb(PORT_EC_CMD)&0x02) );
	return true;
}

bool _writeEC(unsigned char bAddr,unsigned char bData)
{
	/* reset ouput buffer */
	cleanOutputBuffer();

	/* Data : command */
	outb(bAddr,PORT_EC_CMD);
	waitForWriteFinished();

	/* Data : data */
	outb(bData, PORT_EC_DATA);
	waitForWriteFinished();

	return true;
}

bool _writeECSpace(unsigned char bAddr,unsigned char bData)
{
	/* reset ouput buffer */
	cleanOutputBuffer();

	/*command  */
	outb(CMD_WR, PORT_EC_CMD);
	waitForWriteFinished();

	/* Data : address */
	outb(bAddr, PORT_EC_DATA);
	waitForWriteFinished();

	/* Data : data */
	outb(bData, PORT_EC_DATA);
	waitForWriteFinished();

	return true;
}


unsigned char _readEC(unsigned char bAddr)
{
	/* reset ouput buffer */
	cleanOutputBuffer();

	outb(bAddr, PORT_EC_CMD);

	/* reset ouput buffer */
	waitForWriteFinished();

	waitForDataReady();

	return inb(PORT_EC_DATA);

}

unsigned char _readECSpace(unsigned char bAddr)
{
	/* reset ouput buffer */
	cleanOutputBuffer();

	/*command  */
	outb((unsigned char)CMD_RD, PORT_EC_CMD);
	waitForWriteFinished();

	/* Data : address */
	outb((unsigned char)bAddr, PORT_EC_DATA);
	waitForWriteFinished();

	waitForDataReady();

	return inb(PORT_EC_DATA);
}

bool writeEC(unsigned char bAddr,unsigned char bData)
{
	bool ret = true;
	switch (bAddr)
	{
	case 0x10:
	case 0x11:
	case 0x12:
	case 0x16:
	case 0x15:
	case 0x14:
	case 0x3a:
	case 0x92:
	case 0x93:
	case 0x3D:
	case 0x46:
//	case 0x5C:
		_writeECSpace(bAddr, bData);
		break;
	default:
		_writeEC(bAddr, bData);
		break;
	}
	return ret;
}
unsigned char readEC(unsigned char bAddr)
{
	unsigned char bData;

	switch(bAddr)
	{
	case 0x1c:
	case 0x2c:
	case 0x36:
	case 0x38:
	case 0x1a:
	case 0x98:
	case 0x97:
	case 0x92:
	case 0x93:
	case 0x3D:
	case 0x46:
	case 0x3B:
		bData = _readECSpace(bAddr);
		break;
	case 0x00:
	case 0x01:
	case 0x02:
	case 0x03:
	case 0x04:
	case 0x05:
	case 0x06:
	case 0x07:
	case 0x08:
	case 0x09:
	case 0x0a:
	case 0x0b:
		bData = _readECSpace(bAddr);
		if((bData < 0x20) || (bData > 0x7a))
		{
			bData = 0x5F;
		}
		break;
	case 0x10:
	case 0x12:
	case 0x14:
	case 0x16:
	case 0x18:
	case 0x1E:
	case 0x20:
	case 0x22:
	case 0x24:
	case 0x26:
	case 0x28:
	case 0x2E:
	{
		unsigned short temp;
		temp = _readECSpace(bAddr);
		temp <<= 8;
		temp |= _readECSpace(bAddr+1);
		if (temp > MAX_BATTERY_CAPACITY)
		{
			switch(bAddr)
			{
			case 0x12:
			case 0x16:
			case 0x22:
			case 0x26:
				temp=0;
				break;
			}
		}
		bData = (unsigned char)temp;
		break;
	}
	default:
		bData = _readEC(bAddr);
		break;
	}

	return bData;
}
//----------------------------------------------------------------------------------------------------------------------------------------------------

ssize_t x86_write(struct file *filp,const char __user * buffer, size_t length, loff_t *offset)
{    
	char *after;
	unsigned char data;
	unsigned char addr;
	char str[5];
	char cBuf[10];

	copy_from_user(cBuf,buffer,length);
	switch(length)
	{
	case 9:
		strncpy(str,cBuf,4);
		str[4] = '\0';
		addr = simple_strtoul(str,&after,0);
		data = simple_strtoul(cBuf+4,&after,0);
		writeEC(addr,data);
		break;
	case 5:
		addr = simple_strtoul(cBuf,&after,0);
		gcData = readEC(addr);
		gfDataReady =true;
		break;
	default:
		return 0;
	}
	return length;
};


ssize_t x86_read(struct file *filp, char __user * buffer, size_t length, loff_t *offset)
{	
	char cBuf[10];
	if (gfDataReady)
	{
		snprintf(cBuf,3,"%02x",gcData);
		copy_to_user(buffer,cBuf,3);
		gfDataReady = false;
		return 3;
	}

	return 0;
};

static char val[2] = {0, 0};
long x86_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
	long ret=0;
	unsigned char addr;
	unsigned char cData;
	
	if(_IOC_TYPE(cmd) != IOC_MAGIC){
		return -ENOTTY;
	}
	if(_IOC_NR(cmd) > IOC_MAXNR){
		return -ENOTTY;
	}
	
	switch(cmd) {
		case SETNUM:
			//copy data from arg(user) to val(kernel)
			if(copy_from_user(val,(int __user *)arg,2)){
				return -1;
			}
			addr = (unsigned char) val[1];//(val >> 8);
			cData = (unsigned char) val[0];//(val & 0xff);
			writeEC(addr, cData);
			//printk("%s():get val from user: [0] = 0x%02x, [1] = 0x%02x\n", 
			//	__FUNCTION__, (int)val[0], (int)val[1]);
			break;

		case GETNUM:
			if(copy_from_user(val,(int __user *)arg,1)){
				return -1;
			}
			//copy data from val to arg
			addr = (unsigned char) val[0];
			cData = readEC(addr);
			val[0] = (int)cData;
			if(copy_to_user((int __user *)arg,val,1)){
				return -1;
			}

			//printk("%s():set val to %d\n",__FUNCTION__, val[0]);
			break;
	}
	 
	return ret;
}

int x86_release(struct inode *inode, struct file *filp)
{
	//printk("x86_RW: release\n");
	return 0;
};


int x86_open(struct inode *inode, struct file *filp)
{
	//printk(KERN_WARNING "x86_RW: Open \n");
	return 0;
};



//-------------------------------------------------------------------------------------------------------------------------------------------------

struct file_operations EC_fops = 
{
	open: x86_open,
	write: x86_write,
	read: x86_read,
	unlocked_ioctl: x86_ioctl,
	release: x86_release,
};

static int EC_init(void)
{
	int resultState=0;

	printk("x86_rw module init\n");		

	/*  Use fixed Major number*/
	devno = MKDEV( DEV_MAJOR, scull_minor);
	resultState = register_chrdev_region(devno, 2, "x86_RW");
	devnoflg = 1;
		
	if (resultState < 0) {
		printk(KERN_WARNING "scull: can't get major %d\n", scull_major);
		devnoflg = 0;
		return resultState;
	}
	
	ec_cdev = cdev_alloc();

	if(NULL == ec_cdev){
		printk("can't request the memory!\n");
	}

	ec_cdev->owner = THIS_MODULE;
	ec_cdev->ops = &EC_fops;
	resultState = cdev_add(ec_cdev,devno,1);

	if(resultState < 0)
	{
		printk("can't add cdev.2\n");
		adddevflg = 0;
	}
	else
	{
		adddevflg = 1;
	}

	class_EC = class_create( THIS_MODULE, "x86_RW");
	device_EC = device_create(class_EC, NULL, MKDEV(DEV_MAJOR,scull_minor), NULL, "x86_RW");

#if 1
	ec_button_cdev = cdev_alloc();
	if(NULL == ec_button_cdev){
		printk("can't request the memory!\n");
	}
	cdev_init(ec_button_cdev, &EC_fops);
	cdev_add(ec_button_cdev, MKDEV(DEV_MAJOR, scull_minor+1), 1);

	//device_EC_button = device_create(class_EC, NULL, MKDEV(DEV_MAJOR,scull_minor+1), NULL, "npce985la0dx");/*_button*/
#endif
	//signal(SIGALRM, sigalrm_fn);
	
	printk("init success\n");

	return resultState;
}


static void EC_exit(void)
{
	if(adddevflg)
	{
		printk("adddev\n");
		cdev_del(ec_cdev);
		cdev_del(ec_button_cdev);
		adddevflg = 0;
	}

	if(devnoflg)
	{
		printk("devnoflg\n");
		cdev_del(ec_cdev);
		cdev_del(ec_button_cdev);
		devno=MKDEV(DEV_MAJOR,0);
		device_destroy(class_EC,devno);
//		devno=MKDEV(DEV_MAJOR,1);
//		device_destroy(class_EC,devno);
		class_destroy(class_EC);
		unregister_chrdev_region(devno,2);
		devno = 0;
		devnoflg = 0;
	}
	printk("back here and\n");	   
	printk("Goodbye, x86_RW\n");//KERN_ALER
}

module_init(EC_init);
module_exit(EC_exit);
MODULE_LICENSE("Dual BSD/GPL");
